#!/usr/bin/env python3
"""
Converts Private-Radar KML files to GPX files usable in Ayvri

:author: Thomas Calmant
:copyright: Copyright 2022, Thomas Calmant
:license: Apache License 2.0
:version: 0.0.1
..
    Copyright 2022 Thomas Calmant
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import logging
import pathlib
import sys
from typing import List, Optional, cast

import gpxpy
import gpxpy.gpx

from . import __docformat__, __version__, __version_info__
from .beans import AbstractInputHandler
from .kml import KmlInputHandler
from .private_radar import PrivateRadarHandler


def main(args: Optional[List[str]] = None) -> int:
    """
    Script entry point
    """
    # Known handlers
    input_handlers: List[AbstractInputHandler] = [
        KmlInputHandler(),
        PrivateRadarHandler(),
    ]

    # Setup the arguments parser
    parser = argparse.ArgumentParser("kml2gpx")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Set logger to DEBUG level"
    )
    parser.add_argument(
        "-o", "--output", type=pathlib.Path, help="Output GPX file"
    )

    subparsers = parser.add_subparsers(
        description="Input-based arguments", required=True
    )
    for handler in input_handlers:
        sub_parser = subparsers.add_parser(
            handler.get_id(), description=handler.get_description()
        )
        sub_parser.set_defaults(handler=handler)
        handler.register_arguments(sub_parser)

    # Parse arguments
    options = parser.parse_args(args)

    logging.basicConfig(
        level=logging.DEBUG if options.verbose else logging.ERROR
    )

    # Get the handler
    handler: AbstractInputHandler = options.handler
    try:
        rc = handler.check_arguments(options)
        if rc is not None:
            return rc
    except (ValueError, TypeError) as ex:
        print(ex, file=sys.stderr)
        return 1

    # Compute output file
    output = cast(Optional[pathlib.Path], options.output)
    if output is None:
        output = handler.get_default_output_path()

    if output is None:
        output = pathlib.Path("output.gpx")

    # Prepare the root GPX structure
    gpx = gpxpy.gpx.GPX()
    gpx.tracks.append(handler.to_gpx())

    # Write the GPX file
    with open(output, "w", encoding="utf-8") as gpx_fd:
        gpx_fd.write(gpx.to_xml())

    print(output, "written successfully.")
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
Utility module to parse KML files

:author: Thomas Calmant
:copyright: Copyright 2022, Thomas Calmant
:license: Apache License 2.0
:version: 0.0.1
..
    Copyright 2022 Thomas Calmant
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import pathlib
from datetime import datetime
from typing import Dict, List, Optional, Union, cast

import gpxpy
import gpxpy.gpx
import lxml.etree

from . import __docformat__, __version__, __version_info__
from .beans import AbstractInputHandler, TrackPoint


class KmlParser:
    def __init__(self) -> None:
        self.kml: Optional[lxml.etree._Element] = None

    def load(self, in_path: pathlib.Path) -> None:
        with open(in_path, "rb") as fd:
            kml: lxml.etree._Element = lxml.etree.parse(fd, None).getroot()

            root_ns = kml.nsmap[kml.prefix]
            if root_ns != "http://www.opengis.net/kml/2.2":
                raise TypeError("Not a valid/supported KML file")

            self.kml = kml

    def get_coordinates(
        self, layer_name: str = "Altitude",
    ) -> List[TrackPoint]:
        if self.kml is None:
            raise IOError("KML data not loaded yet")

        nsmap: Dict[str, str] = self.kml.nsmap
        for placemark in self.kml.findall("Placemark", nsmap):
            placemark = cast(lxml.etree._Element, placemark)
            name = placemark.findtext("name", None, nsmap)
            if name == layer_name:
                break
        else:
            raise KeyError(layer_name)

        alt_mode = placemark.findtext("LineString/altitudeMode", None, nsmap)
        if alt_mode != "absolute":
            raise ValueError(f"Altitude mode is not absolute: {alt_mode}")

        nodes = placemark.findall("LineString/coordinates", nsmap)
        if not nodes:
            raise ValueError(f"No nodes found in layer {layer_name}")

        coords_list_raw: str = nodes[0].text
        coords: List[TrackPoint] = []
        for row in coords_list_raw.split(None):
            row = row.strip()
            if not row:
                # Empty line
                continue

            str_values = row.strip().split(",")
            if len(str_values) != 3:
                raise ValueError("Not a lon,lat,alt row: ", row)

            lon = float(str_values[0])
            lat = float(str_values[1])
            alt = float(str_values[2])
            trk_pt = TrackPoint(lon, lat, alt)
            coords.append(trk_pt)

        return coords


class KmlInputHandler(AbstractInputHandler):
    """
    Handles KML file input
    """

    def __init__(self) -> None:
        self.in_files: List[pathlib.Path] = []
        self.layer: str = "Altitude"
        self.start_times: Union[List[None], List[datetime]] = []
        self.end_times: Union[List[None], List[datetime]] = []

    def get_id(self) -> str:
        """
        Returns the ID of the handler, used as sub-parser name
        """
        return "kml"

    def get_description(self) -> str:
        """
        One-line description of the handler
        """
        return "KML input handler"

    def register_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Register CLI arguments
        """
        parser.add_argument(
            "-i",
            "--input",
            required=True,
            type=pathlib.Path,
            action="extend",
            nargs="+",
            help="Input KML file(s)",
        )
        parser.add_argument(
            "--start",
            action="extend",
            nargs="*",
            help="Start time of the files",
        )
        parser.add_argument(
            "--end", action="extend", nargs="*", help="End time of the files"
        )
        parser.add_argument(
            "--layer",
            default="Altitude",
            help="KML layer to extract (Altitude by default)",
        )

    def check_arguments(self, options: argparse.Namespace) -> None:
        """
        Checks input arguments

        :raise ValueError: Invalid argument
        """
        # Cast arguments
        self.layer = cast(str, options.layer)
        if not self.layer:
            raise ValueError("No KML layer given.")

        self.in_files = cast(List[pathlib.Path], options.input)
        raw_start_times = cast(List[str], options.start or [])
        raw_end_times = cast(List[str], options.end or [])

        # Check files
        if not self.in_files:
            raise ValueError("No input file given.")

        for file in self.in_files:
            if not file.exists():
                raise ValueError(f"File not found: {file}")

        # Check times
        if len(raw_start_times) != len(raw_end_times):
            raise ValueError(
                "There must be the same number of start and end times"
            )

        nb_files = len(self.in_files)
        if (
            len(raw_start_times) != nb_files
            and len(raw_start_times) != nb_files
        ):
            raise ValueError("There must be as many start times as input files")

        if not raw_start_times:
            # No times given
            self.start_times = [None] * nb_files
            self.end_times = [None] * nb_files
        else:
            # Parse times
            self.start_times = [
                datetime.fromisoformat(raw) for raw in raw_start_times
            ]
            self.end_times = [
                datetime.fromisoformat(raw) for raw in raw_end_times
            ]

    def get_default_output_path(self) -> Optional[pathlib.Path]:
        """
        Returns the name of the default output path, if any
        """
        return self.in_files[0].with_suffix(".gpx")

    def to_gpx(self) -> gpxpy.gpx.GPXTrack:
        """
        Converts the input to a GPX Track
        """
        gpx_track = gpxpy.gpx.GPXTrack()

        # Parse files
        for in_file, start_time, end_time in zip(
            self.in_files, self.start_times, self.end_times
        ):
            # Parse the KML file
            kml = KmlParser()
            kml.load(in_file)
            file_coords = kml.get_coordinates(self.layer)

            # Update times
            if start_time is not None and end_time is not None:
                self.set_times(file_coords, start_time, end_time)

            # Create the GPX segment
            gpx_segment = gpxpy.gpx.GPXTrackSegment()
            gpx_segment.points = [trk.to_gpx() for trk in file_coords]
            gpx_track.segments.append(gpx_segment)

        return gpx_track

    @staticmethod
    def set_times(
        coords: List[TrackPoint], start_time: datetime, end_time: datetime
    ):
        """
        Forces the update of the time field in the given track points, evenly
        spread between the start and end time
        """
        time = start_time
        time_delta = (end_time - start_time) / len(coords)
        for coord in coords:
            coord.time = time
            time += time_delta



#!/usr/bin/env python3
"""
Definition of beans used across the project

:author: Thomas Calmant
:copyright: Copyright 2022, Thomas Calmant
:license: Apache License 2.0
:version: 0.0.1
..
    Copyright 2022 Thomas Calmant
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import pathlib
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

import gpxpy.gpx

from . import __docformat__, __version__, __version_info__

__all__ = ["TrackPoint", "AbstractInputHandler"]


@dataclass
class TrackPoint:
    longitude: float
    latitude: float
    altitude: float
    time: Optional[datetime] = None

    def to_gpx(self):
        return gpxpy.gpx.GPXTrackPoint(
            self.latitude, self.longitude, self.altitude, self.time
        )


class AbstractInputHandler:
    """
    Abstract class of input handlers
    """

    def get_id(self) -> str:
        """
        Returns the ID of the handler, used as sub-parser name
        """
        raise NotImplementedError

    def get_description(self) -> str:
        """
        One-line description of the handler
        """
        return ""

    def register_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Register CLI arguments
        """

    def check_arguments(self, options: argparse.Namespace) -> Optional[int]:
        """
        Checks input arguments

        :return: None if all good, an exit code if an argument was handled
        :raise ValueError: Invalid argument
        """

    def get_default_output_path(self) -> Optional[pathlib.Path]:
        """
        Returns the name of the default output path, if any
        """
        return None

    def to_gpx(self) -> gpxpy.gpx.GPXTrack:
        """
        Converts the input to a GPX Track
        """
        raise NotImplementedError



#!/usr/bin/env python3
"""
Utility module to access Private Radar flights information

:author: Thomas Calmant
:copyright: Copyright 2022, Thomas Calmant
:license: Apache License 2.0
:version: 0.0.1
..
    Copyright 2022 Thomas Calmant
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import configparser
import logging
import pathlib
from configparser import SafeConfigParser
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Union, cast
from urllib.parse import urljoin

import gpxpy.gpx
import requests

from . import __docformat__, __version__, __version_info__
from .beans import AbstractInputHandler, TrackPoint


@dataclass
class PRFlight:
    """
    Description of a Private Radar flight
    """

    id: int
    """ Private Radar ID of the flight """

    registration: str
    """ Airplane registration number """

    from_icao: str
    """ ICAO code of the departure airport """

    to_icao: str
    """ ICAO code of the arrival airport """

    flight_type: str
    """ Kind of flight: solo, solo_sup, instr, ... """

    start: datetime
    """ Engine start time """

    end: datetime
    """ Engine shut down time """

    crew: List[str]
    """ List of names of crew members """

    starred: bool = False
    """ Starred flight """

    @classmethod
    def parse(cls, data: Dict[str, Any]) -> "PRFlight":
        """
        Constructs a Private Radar flight from its JSON description
        """
        # Parse dates
        start = datetime.fromisoformat(data["dt_start"])
        end = datetime.fromisoformat(data["dt_terminated"])

        # Parse the crew
        crew = [
            f"{member['firstname']} {member['lastname']}"
            for member in data["crew"]
        ]

        # Get the rest of details as is
        return cls(
            data["id"],
            data["registration"],
            data["from"],
            data["to"],
            data["flight_type"],
            start,
            end,
            crew,
            data["starred"],
        )


class PrivateRadar:
    """
    API to get Private Radar data
    """

    def __init__(self, base_url: str, auth: str) -> None:
        """
        :param base_url: Private Radar base URL
        :param auth: Private Radar authentication token
        """
        self.base_url = base_url
        self.auth = auth
        self.logger = logging.getLogger(__name__)

    def list_flights(
        self, nb_flights: int = 100, flight_id: Optional[int] = None
    ) -> List[PRFlight]:
        """
        Lists flights from Private Radar

        :param nb_flights: Maximum number of flights to return
        :param flight_id: Return details for that specific flight
        """
        filter_flight_id: str = ""
        filter_type: int = 1

        if flight_id is not None:
            # Select a specific flight
            filter_flight_id = str(flight_id)
            filter_type = 5

        # Get last flights
        params = {
            "filter6": {
                "type": filter_type,
                "page": 1,
                "unit": 0,
                "nbFlights": nb_flights,
                "flightTypes": "solo,solo_sup,solo_only,instr,mcc,exam,tourist,discovery,check,ferry,trip,rental,tow,aoc",
                "aircraft": 0,
                "crew": 0,
                "showRoute": True,
                "showTotalTime": True,
                "showBlockTime": True,
                "showAirtime": True,
                "showCrew": True,
                "showMap": True,
                "hideTaxi": True,
                "flightIds": filter_flight_id,
                "nbUnits": nb_flights,
                "date": 0,
                "dateFrom": 0,
                "dateTo": 0,
            }
        }

        # Query the list
        response = requests.post(
            urljoin(self.base_url, "/prwsw/flight/getFlightsFilter6"),
            headers={
                "Authorization": self.auth,
                "Accept": "application/json, text/plain, */*",
            },
            json=params,
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as ex:
            self.logger.error("Error getting the list of flights: %s", ex)
            raise ex

        # Parse flights
        result = cast(Dict[str, Any], response.json())

        if result["status"] != "success":
            error = result.get("error", "n/a")
            self.logger.error(f"Error retrieving flights: {error}")
            raise IOError(f"Error retrieving flights: {error}")

        with open("out.json", "w") as fd:
            fd.write(response.text)
        flights_json: List[Dict[str, Any]] = response.json()["flight_list"]
        return [PRFlight.parse(flight) for flight in flights_json]

    def get_flight(self, flight_id: int) -> PRFlight:
        """
        Returns the description of the given flight

        :param flight_id: Private Radar flight ID
        :return: The description of the flight
        """
        if flight_id >= 0:
            return self.list_flights(100, flight_id=flight_id)[0]
        else:
            nb_flights = abs(flight_id)
            return self.list_flights(nb_flights)[abs(flight_id) - 1]

    def flight_path(self, flight: Union[int, PRFlight]) -> List[TrackPoint]:
        """
        Retrieves the path of the given Private Radar path
        """
        if isinstance(flight, PRFlight):
            flight_id = flight.id
        elif isinstance(flight, int):
            flight_id = flight
        else:
            raise TypeError(f"Invalid flight ID: {type(flight).__name__}")

        # Get the flight
        response = requests.post(
            urljoin(self.base_url, "/prwsw/flight/get_path"),
            headers={
                "Authorization": self.auth,
                "Accept": "application/json, text/plain, */*",
            },
            json={"id": flight_id},
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as ex:
            self.logger.error(
                "Error getting path for flight %d: %s", flight_id, ex
            )
            raise ex

        # Check its content
        flight_details: Dict[str, Any] = response.json()
        if flight_details.get("status") != "success":
            self.logger.error("Error getting flight %d", flight_id)
            self.logger.debug("Flight details:\n%s", flight_details)
            raise ValueError("Server-side error getting the flight")

        # Parse the flight track
        path: List[Dict[str, Any]] = flight_details["flight_profile"]
        return [
            TrackPoint(
                node["lon"],
                node["lat"],
                node["alt_m"],
                # Convert Private Radar timestamps
                datetime.fromtimestamp(node["time"] * 10 ** -3),
            )
            for node in path
        ]


class PrivateRadarHandler(AbstractInputHandler):
    """
    Handles flights from PrivateRadar
    """

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.api: Optional[PrivateRadar] = None
        self.flight: Optional[PRFlight] = None

    def get_id(self) -> str:
        """
        Returns the ID of the handler, used as sub-parser name
        """
        return "private-radar"

    def get_description(self) -> str:
        """
        One-line description of the handler
        """
        return "Load flights from Private Radar"

    def register_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Register CLI arguments
        """
        group = parser.add_argument_group("Configuration")
        group.add_argument(
            "-c",
            "--conf",
            type=pathlib.Path,
            help="Path to the configuration file",
        )
        group.add_argument("--url", help="Base URL of Private Radar")
        group.add_argument("--auth", help="Authentication token")

        group = parser.add_argument_group("Flights list")
        group.add_argument(
            "-l",
            "--list",
            action="store_true",
            help="List available flights and exit.",
        )
        group.add_argument(
            "-n",
            "--nb",
            default=50,
            type=int,
            help="Number of available flights to list",
        )

        parser.add_argument(
            "-f",
            "--flight",
            type=int,
            help="ID of the flight to extract (or -1 for last flight, ...)",
        )

    def _find_conf(self) -> Optional[pathlib.Path]:
        """
        Tries to find a path for the configuration file in well-known folders
        """
        for folder in (
            pathlib.Path("."),
            pathlib.Path("~"),
            pathlib.Path("~/.conf"),
        ):
            folder = folder.expanduser()
            for name in ("kml2gpx.ini", "private_radar.ini"):
                conf_path = folder / name
                if conf_path.exists():
                    return conf_path

        return None

    def check_arguments(self, options: argparse.Namespace) -> Optional[int]:
        """
        Checks input arguments

        :raise ValueError: Invalid argument
        """
        url: Optional[str] = None
        token: Optional[str] = None

        # Try to load the configuration from a file
        conf_path = cast(Optional[pathlib.Path], options.conf)
        if conf_path is None:
            conf_path = self._find_conf()

        if conf_path is not None:
            if not conf_path.exists():
                raise ValueError(f"Configuration file not found: {conf_path}")

            conf_path = conf_path.absolute()
            config = SafeConfigParser()
            with open(conf_path, "r") as fd:
                self.logger.debug("Reading configuration file: %s", conf_path)
                config.read_file(fd)

            try:
                url = config.get("PRIVATE_RADAR", "url")
                token = config.get("PRIVATE_RADAR", "token")
            except configparser.NoSectionError as ex:
                raise ValueError(f"Invalid configuration file: {ex}")

        # Override with arguments, if any
        if options.url:
            url = cast(Optional[str], options.url)

        if options.auth:
            token = cast(Optional[str], options.auth)

        if not url:
            raise ValueError("Missing Private Radar URL configuration")

        if not token:
            raise ValueError(
                "Missing Private Radar authentication configuration"
            )

        # Prepare the API
        self.logger.debug("Using Private Radar URL: %s", url)
        self.api = PrivateRadar(url, token)

        if options.list:
            # List flights and exit
            return self.print_flights(options.nb)
        else:
            flight_id = cast(Optional[int], options.flight)
            if flight_id is None:
                raise ValueError("A flight ID is required")

            # Load the flight
            self.flight = self.api.get_flight(flight_id)
            print("Loaded flight:")
            self.print_flight(self.flight)

    def get_default_output_path(self) -> Optional[pathlib.Path]:
        """
        Returns a default name for the output GPX file
        """
        if self.flight is None:
            return None

        return pathlib.Path(
            f"{self.flight.start.strftime('%Y-%m-%d')}-"
            f"{self.flight.from_icao}-"
            f"{self.flight.to_icao}.gpx"
        )

    def to_gpx(self) -> gpxpy.gpx.GPXTrack:
        """
        Converts the input to a GPX Track
        """
        if self.api is None:
            raise ValueError("Private Radar API not set up")

        if self.flight is None:
            raise ValueError("Flight not loaded.")

        gpx_segment = gpxpy.gpx.GPXTrackSegment()
        gpx_segment.points = [
            point.to_gpx() for point in self.api.flight_path(self.flight)
        ]

        gpx_track = gpxpy.gpx.GPXTrack()
        gpx_track.segments.append(gpx_segment)
        return gpx_track

    def print_flights(self, nb_flights: int) -> int:
        """
        Prints out flights visible in Private Radar

        :param nb_flights: Number of flights to show
        :return: 0 on success, 1 on error
        """
        if self.api is None:
            self.logger.error("Private Radar API not set up")
            return 1

        for flight in self.api.list_flights(nb_flights):
            self.print_flight(flight)

        return 0

    @staticmethod
    def print_flight(flight):
        start = flight.start.isoformat(" ") if flight.start else "n/a"
        end = flight.end.isoformat(" ") if flight.end else "n/a"

        sep = "---" if not flight.starred else "==="
        star = " (*)" if flight.starred else ""

        print(f"{sep} Flight #{flight.id}{star} {sep}")
        print("* From.:", flight.from_icao)
        print("* To...:", flight.to_icao)
        print("* Crew.:", ", ".join(flight.crew))
        print("* Type.:", flight.flight_type)
        print("* Start:", start)
        print("* End..:", end)
        print()



#!/usr/bin/env python3
"""
KML2GPX project: converts KML files or flights from Private Radar to GPX

:author: Thomas Calmant
:copyright: Copyright 2022, Thomas Calmant
:license: Apache License 2.0
:version: 0.0.1
..
    Copyright 2022 Thomas Calmant
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Module version
__version_info__ = (0, 0, 1)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"
